# Copyright (C) 2025 National Institute of Advanced Industrial Science and Technology (AIST)
# SPDX-License-Identifier: MIT

from aiaccel.torch.functional.linear_sum_assignment import linear_sum_assignment

__all__ = [
    "linear_sum_assignment",
]
