# Copyright (C) 2025 National Institute of Advanced Industrial Science and Technology (AIST)
# SPDX-License-Identifier: MIT

from aiaccel.torch.lightning.datamodules.single_datamodule import SingleDataModule

__all__ = ["SingleDataModule"]
